'use strict';
//
// ウィンドウ制御
//
var W = W || { // 名前空間
	// ウィンドウ情報取得
	getWin(id) {
		return W._win[id];
	},

	// カレントウィンドウ情報取得
	getCrntWin() {
		let crntWin = null;
		let crntLv = 0;
		let wins =  W._root.pane.getElementsByTagName('w-win-area');
		for (let win of wins) {
			let rect = win.getBoundingClientRect(); // マウスチェック
			if (W._mouseX < rect.left) continue;
			if (W._mouseX > rect.right) continue;
			if (W._mouseY < rect.top) continue;
			if (W._mouseY > rect.bottom) continue;

			let lv = 0;
			let elem = win;
			while (elem != null) { // 階層レベル(1-n)を取得
				if (elem.tagName == 'W-WIN-AREA') lv++;
				elem = elem.parentNode;
			}
			if (lv < crntLv) continue; // 子階層を優先
			crntLv = lv; // カレントウィンドウの階層レベルを設定
			crntWin = win; // カレントウィンドウの設定
		}
		let id = crntWin.id.slice(6);
		return W._win[id];
	},

	// 要素からウィンドウ情報取得
	getElemWin(elem) {
		let id = '';
		while (elem != null) {
			if (elem.tagName == 'W-WIN') {
				id = elem.id;
				break;
			} else if (elem.tagName == 'W-WIN-AREA') {
				id = elem.id.slice(6);
				break;
			} else if (elem.tagName == 'W-PANE-AREA') {
				id = elem.id.slice(7);
				break;
			}
			elem = elem.parentNode;
		}
		return W._win[id];
	},

	// ウィンドウ制御 内部プロパティ
	_root: null, // MainWinクラス
	_win: new Object(), // Winクラス
	_observer: null, // イベント監視
	_resizeOb: null, // リサイズイベント監視
	_mouseX: 0, // マウスX座標
	_mouseY: 0, // マウスY座標
	_dragX: 0, // ドラッグ中のマウスX座標
	_dragY: 0, // ドラッグ中のマウスY座標
	_dragWin: null, // ドラッグ対象のウィンドウ要素<w-area>
	_dragElem: null, // ドラッグ要素
	_dragDist: 0, // ドラッグ要素<w-slide>の移動距離
	_crntWin: null, // カレントウィンドウ要素
	_Dialog: null, // ダイアログ要素

	// 初期処理
	_init(main) {
		W._root = main;

		W._resizeOb = new ResizeObserver(entries => { // サイズ変更監視
			for (let entry of entries) {
				let win = W._win[entry.target.id.slice(6)];
				win.resize(entry);
			}
		});

		W._root.pane.onmousemove = function(event) {
			W._mouseX = event.clientX;
			W._mouseY = event.clientY;

			if (W._dragWin != null) { // ドラッグ要素の移動
				if (W._dragWin.parentNode.getAttribute('w-rc') == 'row') { // 左右
					if (W._dragWin.getAttribute('w-dir') == 'right-bottom') { // 右
						let d = event.screenX - W._dragDist;
						W._dragElem.style.left = d + 'px';
					} else { // 左 (left-top)
						let d = W._dragDist - event.screenX;
						W._dragElem.style.right = d + 'px';
					}
				} else { // 上下 (column)
					if (W._dragWin.getAttribute('w-dir') == 'right-bottom') { // 下
						let d = event.screenY - W._dragDist;
						W._dragElem.style.top = d + 'px';
					} else { // 上 (left-top)
						let d = W._dragDist - event.screenY;
						W._dragElem.style.bottom = d + 'px';
					}
				}
			}
		}
		W._root.pane.onmouseup = W._dragEnd;
		W._root.pane.onmouseleave = W._dragCancel;
		W._root.pane.ondragstart = function(event) {
			event.preventDefault(); // デフォルトのドラッグ処理を無効化
			return false;
		}
		W._root.pane.onselectstart = function(event) {
			if (W._dragWin != null) { // ドラッグ中は文字列選択不可
				event.preventDefault();
				return false;
			}
		}
		document.onkeydown = function(event) {
			if (W._Dialog != null) { // Dialogを表示中
				W._Dialog.keydown(event);
			} else {
				let win = W.getCrntWin(); // カレントウィンドウ情報取得
				win.keydown(event)
			}
		}
	},

	// 初期処理
	_term() {
		W._resizeOb.disconnect();
		W._root = null;
	},

	// 幅変更ボタンドラッグ開始
	_dragStart(event) {
		if (this.tagName != 'W-SLIDE') return;
		W._dragWin = this.parentNode; // w-win-area
		W._dragX = event.screenX;
		W._dragY = event.screenY;
		W._dragElem = document.createElement('w-slide');
		W._dragWin.appendChild(W._dragElem);
		if (W._dragWin.parentNode.getAttribute('w-rc') == 'row') { // 左右
			if (W._dragWin.getAttribute('w-dir') == 'right-bottom') { // 右
				W._dragDist = Math.floor(W._dragX + W._dragElem.clientWidth / 2);
				W._root.pane.style.cursor = 'w-resize';
			} else { // 左 (left-top)
				W._dragDist = Math.floor(W._dragX - W._dragElem.clientWidth / 2);
				W._root.pane.style.cursor = 'e-resize';
			}
		} else { // 上下 (column)
			if (W._dragWin.getAttribute('w-dir') == 'right-bottom') { // 下
				W._dragDist = Math.floor(W._dragY + W._dragElem.clientHeight / 2);
				W._root.pane.style.cursor = 's-resize';
			} else { // 上 (left-top)
				W._dragDist = Math.floor(W._dragY - W._dragElem.clientHeight / 2);
				W._root.pane.style.cursor = 'n-resize';
			}
		}
	},

	// 幅変更ボタンドラッグ終了
	_dragEnd(event) {
		if (W._dragWin == null) return;
		let f = 0;
		if (W._dragWin.parentNode.getAttribute('w-rc') == 'row') { // 左右
			if (W._dragWin.getAttribute('w-dir') == 'right-bottom') { // 右
				f = W._dragWin.clientWidth + W._dragX - event.screenX;
			} else { // 左 (left-top)
				f = W._dragWin.clientWidth - W._dragX + event.screenX;
			}
		} else { // 上下 (column)
			if (W._dragWin.getAttribute('w-dir') == 'right-bottom') { // 下
				f = W._dragWin.clientHeight + W._dragY - event.screenY;
			} else { // 上 (left-top)
				f = W._dragWin.clientHeight - W._dragY + event.screenY;
			}
		}
		if (f < 50) f = 50; // 最小ウィンドウサイズ
		W._dragWin.style.flex = '0 1 ' + f + 'px';
		W._dragCancel(event);
	},

	// 幅変更ボタンドラッグキャンセル
	_dragCancel(event) {
		if (W._dragElem == null) return;
		W._dragElem.remove();
		W._root.pane.style.cursor = 'default';
		W._dragWin = null;
		W._dragElem = null;
	}
}

// Winクラス
W.Win = class {
	id = ""; // ウィンドウID
	opt = null; // ウィンドウオプション
	win = null; // ウィンドウ要素<w-win>
	area = null; // メインウィンドウ（w-win-area）
	removeCB = null; // ウィンドウ削除時に呼ばれるCB処理
	resizeCB = null; // ウィンドウサイズ変更時に呼ばれるCB処理
	keydownCB = null; // キーボード入力時に呼ばれるCB処理

	constructor(id, opt={}) { // ウィンドウ登録コンストラクタ
		this.id = id;
		this.opt = opt;
		W._win[id] = this; // ウィンドウハッシュの登録

		this.win = document.createElement('w-win');
		this.win.id = this.id;

		this.area = document.createElement('w-win-area');
		this.area.id = "w-win-" + this.id;
		if (typeof opt.flex != 'undefined') this.area.style.flex = opt.flex;
		if (typeof opt.order != 'undefined') this.area.style.order = opt.order;
		this.area.appendChild(this.win);
	}

	remove() { // ウィンドウ削除メソッド
		if (typeof this.removeCB == 'function') this.removeCB();
		this._remove();
	}
	_remove() { // ウィンドウ削除内部メソッド
		this.area.remove(); // DOM要素の削除
		delete W._win[this.id]; // ウィンドウハッシュの削除
	}

	resize(entry) { // ウィンドウサイズ変更メソッド
		this._resize(entry);
		if (typeof this.resizeCB == 'function') this.resizeCB(entry);
	}
	_resize(entry) { // サイズ変更内部メソッド
	}

	keydown(event) { // キーボード入力メソッド
		if (typeof this.keydownCB == 'function') this.keydownCB(event);
		this._keydown(event);
	}
	_keydown(event) { // キーボード入力内部メソッド
	}

	// ボタン<w-btn>を追加する
	// inf: ボタン情報
	//  icon: アイコンの種類
	//  str: アイコン文字列
	//  str1: アイコン文字列1
	//  str2: アイコン文字列2
	//  func: ボタンクリック時の処理
	addBtn(inf) {
		let btn = this._createBtn(inf);
		this.area.appendChild(btn);
		return btn;
	}

	// ツールボックス<w-toolbox>を生成する
	// inf[]: ボタン情報の配列
	// opt{}: 補足情報
	//  elem: infがDOMElementのリスト
	createToolbox(inf, opt = {}) {
		let toolbox = document.createElement('w-toolbox');
		for (let i = 0; i < inf.length; i++) {
			if (opt.elem) {
				toolbox.appendChild(inf[i]);
			} else {
				let btn = this._createBtn(inf[i]);
				toolbox.appendChild(btn);
			}
		}
		return toolbox;
	}

	// ツールボックス<w-toolbox>を追加する
	// inf[]: ボタン情報の配列
	addToolbox(inf) {
		let toolbox = this.createToolbox(inf);
		this.area.appendChild(toolbox);
		return toolbox;
	}

	// ボタン<w-btn>を作成する内部メソッド
	// inf: ボタン情報
	_createBtn(inf) {
		let btn = document.createElement('w-btn');
		if (typeof inf.func == 'function') btn.onclick = inf.func;

		let ico = document.createElement('w-icon');
		ico.classList.add(inf.icon);
		if (typeof inf.str != 'undefined') ico.setAttribute('icon-str', inf.str);
		if (typeof inf.str1 != 'undefined') ico.setAttribute('icon-str1', inf.str1 + '\n');
		if (typeof inf.str2 != 'undefined') ico.setAttribute('icon-str2', inf.str2);
		btn.appendChild(ico);
		return btn;
	}

	// _createBtnの外部参照用ラッパー
	createBtn(inf) {
		let btn = this._createBtn(inf);
		return btn;
	}

	// ラベル<w-lbl>を生成する
	//  str: ラベル文字列
	createLbl(str) {
		let lbl = document.createElement('w-lbl');
		lbl.innerHTML = str;
		return lbl;
	}

	// ラベル<w-lbl>を追加する
	//  str: ラベル文字列
	addLbl(str) {
		let lbl = this.createLbl(str)
		this.area.appendChild(lbl);
		return lbl;
	}

	// プルダウンメニューボタン<w-btn class="w-menu"><ul>を追加する
	// inf: ボタン情報
	//  menu[]: メニュー情報の配列
	//   txt: メニュー表示文字列
	//   cmd  メニューコマンド文字列
	//   opt: メニューパラメータ文字列
	//  func: メニュー選択時の処理
	addMenu(inf) {
		let btn = this.createMenu(inf);
		this.area.appendChild(btn);
		return btn;
	}

	// プルダウンメニューボタン<w-btn class="w-menu"><ul>を生成する
	// inf: ボタン情報
	//  menu[]: メニュー情報の配列
	//   txt: メニュー表示文字列
	//   cmd  メニューコマンド文字列
	//   opt: メニューパラメータ文字列
	//  func: メニュー選択時の処理
	createMenu(inf) {
		let btn = this._createBtn({icon: 'menu'});
		btn.className = 'w-menu';
		this._createMenu(this, btn, inf.func, inf.menu);
		return btn;
	}

	// メニューを作成する内部メソッド（再帰メソッド）
	//  win: ウィンドウクラス
	//  elem: メニューの親要素
	//  func: メニュー処理関数
	//  menu[]: メニュー情報
	_createMenu(win, elem, func, menu) {
		let ul = document.createElement('ul');
		elem.appendChild(ul);
		for (let i = 0; i < menu.length; i++) {
			let li = document.createElement('li');
			ul.appendChild(li);
			li.innerHTML = menu[i].txt;
			if (typeof menu[i].cmd != 'undefined') {
				li.setAttribute('w-cmd', menu[i].cmd);
			}
			if (typeof menu[i].opt != 'undefined') {
				li.setAttribute('w-opt', menu[i].opt);
			}
			if (typeof menu[i].menu != 'undefined') {
				this._createMenu(win, li, func, menu[i].menu);
			} else {
				li.onclick = func;
			}
		}
	}

	// 閉じるボタン作成内部メソッド
	_createCloseBtn() {
		let btn = document.createElement('w-close');
		btn.onclick = function() {
			let win = W._win[this.parentNode.id.slice(6)];
			win.remove();
		}
		this.area.appendChild(btn);
		return btn;
	}

	// 幅変更ボタン作成内部メソッド
	_createSlideBar() {
		let bar = document.createElement('w-slide');
		bar.onmousedown = W._dragStart;
		bar.onmouseup = W._dragCancel;
		this.area.appendChild(bar);
		return bar;
	}
}
W.win = function(id, opt={}) { // 小文字で始まるショートカットの定義
	return new W.Win(id, opt);
}

// MainWinクラス
//  id: メインウィンドウのID
//  pid: 親となるブロック要素のID
//  opt: オプション
//   rc: 分割方法 '':自動分割、'row':横分割、'column':縦分割
W.MainWin = class extends W.Win {
	pid = '';	// 親のDOM要素ID
	parent = null; // 親のDOM要素
	subs = null;	// サブウィンドウ配列
	rc = ''; // 分割方法 '':自動分割、'row':横分割、'column':縦分割
	pane = null;

	constructor(id, pid, opt={}) { // エリア登録コンストラクタ
		super(id, opt);
		this.pid = pid;
		this.parent = document.getElementById(this.pid);
		if (this.parent == null) alert('pid(' + this.pid + ') is not found.');
		this.subs = new Array();

		this.pane = document.createElement('w-pane-area');
		this.pane.id = 'w-pane-' + this.id;
		this.pane.appendChild(this.area);
		if (typeof opt.rc != 'undefined') this.rc = opt.rc;
		this.parent.appendChild(this.pane);

		if (W._root == null) { // ウィンドウ制御 初期化処理
			W._init(this);
		}
		W._resizeOb.observe(this.area); // メインのリサイズ処理を登録
	}

	// エリア削除内部メソッド
	_remove() {
		for (let i = this.subs.length - 1; i >= 0; i--) {
			let win = this.subs[i];
			W._resizeOb.unobserve(win.area); // サブのリサイズ処理の解除
			delete W._win[win.id]; // ウィンドウハッシュの削除
			this.subs.splice(i, 1); // Arrayから削除
		}
		this.pane.remove(); // DOM要素の削除
		W._resizeOb.unobserve(this.area); // メインのリサイズ処理の解除

		if (this.id == W._root.id) {
			W._term();
		}
	}

	// 領域(w-pane-area)リサイズ内部メソッド
	_resize(entry) {
		let rc;
		if (this.rc == '') {
			rc = (this.pane.clientWidth > this.pane.clientHeight) ? 'row' : 'column';
		} else {
			rc = this.rc;
		}
		this.pane.setAttribute('w-rc', rc);
	}

	// サブウィンドウリスト追加
	_addSub(sub) {
		this.subs.push(sub); // サブウィンドウリストに追加
	}

	// サブウィンドウリスト削除
	_delSub(sub) {
		for (let i = 0; i < this.subs.length; i++) {
			if (this.subs[i].id != this.id) continue;
			this.subs.splice(i, 1); // Arrayから削除
		}
	}
}
W.mainWin = function(id, pid, opt={}) { // 小文字で始まるショートカットの定義
	return new W.MainWin(id, pid, opt);
}

// subWinクラス（サブウィンドウクラス）
//  id: サブウィンドウのID
//  main: メインウィンドウクラス
//  opt: オプション
//   closeBtn: 閉じるボタン有無 on:有、off:無
//   slideBar: サイズ変更ボタン有無 on:有、off:無
W.SubWin = class extends W.Win {
	constructor(id, main, opt={}) {
		super(id, opt);

		this.main = main;
		this.main.pane.appendChild(this.area);

		let order = getComputedStyle(this.area).getPropertyValue('order');
		if (order >= 0) {
			this.area.setAttribute('w-dir', 'right-bottom');
		} else {
			this.area.setAttribute('w-dir', 'left-top');
		}
		if ((typeof opt.closeBtn == 'undefined') || (opt.closeBtn != 'off')) {
			this._createCloseBtn(); // 閉じるボタン作成
		}
		if ((typeof opt.slideBar == 'undefined') || (opt.slideBar != 'off')) {
			this._createSlideBar(); // 幅変更ボタン作成
		}

		W._resizeOb.observe(this.area); // サブのリサイズ処理を登録
		this.main._addSub(this); // サブウィンドウリストに追加
	}

	// ウィンドウ削除メソッド
	_remove() {
		W._resizeOb.unobserve(this.area); // サブのリサイズ処理の解除
		this.main._delSub(this); // サブウィンドウリストから削除
		this.area.remove(); // DOM要素の削除
		delete W._win[this.id]; // ウィンドウハッシュの削除
	}
}
W.subWin = function(id, main, opt={}) { // 小文字で始まるショートカットの定義
	return new W.SubWin(id, main, opt);
}

// Dialogクラス
//  opt: オプション
//   id: DialogのID
//   class: Dialogの表示タイプ
W.Dialog = class extends W.Win {
	_dialog = null;
	_overlay = null;

	constructor(title, opt={}) {
		if (W._Dialog != null) { return; }	// Dialogを表示中

		let id = 'wDialog';
		if (typeof opt.id != 'undefined') id = opt.id;
		super(id, opt);

		W._Dialog = this;
		this._dialog = document.createElement('w-dialog');
		this._dialog.appendChild(this.area);
		document.body.appendChild(this._dialog);

		this._overlay = document.createElement('w-overlay');
		document.body.appendChild(this._overlay);

		let wTitleArea = document.createElement('w-title-area');
		this.area.appendChild(wTitleArea);

		let wTitle = document.createElement('w-title');
		wTitle.innerHTML = title;
		wTitleArea.appendChild(wTitle);

		let wTitleBtn = document.createElement('w-title-btn');
		wTitleBtn.onclick = function(event) {
			W._Dialog.remove();
		}
		wTitleArea.appendChild(wTitleBtn);
	}

	// ダイアログ削除時の処理
	_remove() {
		this._overlay.remove(); // DOM要素の削除
		this._dialog.remove(); // DOM要素の削除
		delete W._win[this.id]; // ウィンドウハッシュの削除
		W._Dialog = null;
	}

	// キーボード入力時の処理
	_keydown(event) {
		switch (event.keyCode) { // キーコード
		case 27:	// Esc
			this.remove();
			break;
		}
	}
}
W.dialog = function(opt={}) { // 小文字で始まるショートカットの定義
	return new W.Dialog(opt);
}
